<?php
  //=====================================================//
  //
  // ! YABDAB - MAILER CLASS v.4.0.9
  // - MODIFIED: 2019-09-04 10:07:52 AM
  // - author: Mike Yrabedra
  // - (c)2011-2019 Yabdab Inc. All rights reserved.
  //
  //  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
  //  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  //  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  //  THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  //  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  //  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  //  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  //  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  //  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  //
  //=====================================================//
  
  //SMTP needs accurate times, and the PHP time zone MUST be set
  //This should be done in your php.ini, but this is how to do it if you don't have access to that
  if(ini_get('date.timezone') == '') date_default_timezone_set('GMT');
  
  // ! -- LOAD REQUIRED HELPER CLASSES
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR .'Browser.php'); // include browser sniffer lib
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR .'Spreadsheet.php'); // For Google Sheets
  
  use PHPMailer\PHPMailer\PHPMailer;
  use PHPMailer\PHPMailer\Exception;
  
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR .'PHPMailer/Exception.php');
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR .'PHPMailer/PHPMailer.php');
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR .'PHPMailer/SMTP.php');
  
  
  class YDMailer {
    
    //  Set Variables for Class Use...
    //private $recaptcha;
    private $browser;
    private $mysqli;
    private $mysql_insert_id = '';
    private $timestamp;
    
    public $settings = array();
    public $email_content = '';
    public $receipt_content = '';
    public $files = array();
    public $errors = array();
    public $signature_filename = '';
    
    //======================================================================================//
    // + INSTANTIATE OTHER CLASSES
    //======================================================================================//
    
    function __construct() {
      
      $this->browser = new Browser;
      $this->timestamp = time();
      
    }
    
    //=====================================================//
    // ! - HELPER FUNCTIONS
    //=====================================================//
    
    public function recaptchaCheckAnswer()
    {
      // Get a key from http://recaptcha.net/api/getkey
      /*
       $publickey = $this->settings['recaptcha_public_key'];
       $privatekey = $this->settings['recaptcha_private_key'];
       $resp = null;
       $error = null;
       $resp = recaptcha_check_answer($privatekey,$_SERVER["REMOTE_ADDR"],$_POST['recaptcha_challenge_field'],$_POST['recaptcha_response_field']);
       
       if (!$resp->is_valid) {
       # set the error code so that we can display it
       return $resp->error;
       }
       
       return 'good';
       */
      
      $secret = $this->settings['recaptcha_private_key'];
      
      if(isset($_POST['recaptcha_token']))
        $response=$_POST['recaptcha_token'];
      
      if(!$response){
        return 'Please check the the captcha form.';
        exit;
      }
      
      $verify=file_get_contents("https://www.google.com/recaptcha/api/siteverify?secret={$secret}&response={$response}");
      
      $captcha_success=json_decode($verify);
      
      if ($captcha_success->success==false) {
        return "damn bots";
      }
      else if ($captcha_success->success==true) {
        return "good";
      }
      
      
    }
    
    public function processForm()
    {
      
      if(!$this->settings['do_not_send'])
      {
        // send main message
        $this->buildEmailContent();
        
        $send = $this->sendMessage();
        
        if($send != 'good')
          @array_push($this->errors, $send);
        
        // send receipt to submitter
        if($this->settings['send_receipt'])
        {
          $this->buildReceiptContent();
          
          $receipt = $this->sendMessage(true);
          
          if($receipt != 'good')
            @array_push($this->errors, $receipt);
        }
      }
      
      // Save to MySQL?
      if($this->settings['save_to_mysql']){
        $rs = $this->saveToMySQL();
        if($rs != 'good')
          @array_push($this->errors, $rs);
        unset($rs);
      }
      
      // Save to Google Sheets?
      if($this->settings['save_to_google']){
        $rs = $this->saveToGoogle();
        if($rs != 'good')
          @array_push($this->errors, $rs);
        unset($rs);
      }
      
      
      
    }
    
    //=====================================================//
    // processAttachments
    //=====================================================//
    
    public function processAttachments()
    {
      if (isset($_FILES)) {
        
        foreach ($_FILES as $key=>$attachment) {
          
          $attachment_name = $attachment['name'];
          $attachment_size = $attachment["size"];
          $attachment_temp = $attachment["tmp_name"];
          $attachment_type = $attachment["type"];
          $attachment_ext  = explode('.', $attachment_name);
          $attachment_ext  = $attachment_ext[count($attachment_ext)-1];
          
          if ( $attachment['error'] === UPLOAD_ERR_OK && is_uploaded_file($attachment['tmp_name']) ) {
            
            // User set filtering; size, ext, etc...
            if(@stristr($this->settings['allowed_file_types'], $attachment_ext) == FALSE){ // make sure we accet this file type....
              return $this->settings['file_not_allowed'];
            }
            
            // See is file is not too big.
            if($attachment_size > $this->settings['max_attachment_size_bytes']){
              return $this->settings['file_too_big'] ;
            }
            
            // Check if file already exists
            // Save to local folder?
            
            if($this->settings['save_attachment_files']){
              
              $target_file = $this->settings['upload_folder'] . $this->timestamp. '_' .basename( $attachment_name );
              
              if (!copy( $_FILES[$key]['tmp_name'], $target_file )) {
                $error = "Sorry, there was an error saving your file(s).";
              }
              
              // give new location of uploaded file
              array_push( $this->files, array('temp'=> $target_file, 'name' => $this->timestamp. '_' .$attachment_name, 'key' => $key ));
              
            }else{
              
              // use temp location of file
              array_push( $this->files, array('temp'=> $attachment_temp, 'name' => $this->timestamp. '_' .$attachment_name, 'key' => $key ));
              
            }
            
            
          } else if ($attachment['error'] !== UPLOAD_ERR_NO_FILE) {
            
            // ! -- SERVER ERRORS
            switch ($attachment['error']) {
              case UPLOAD_ERR_INI_SIZE:
              case UPLOAD_ERR_FORM_SIZE:
                $error = "File $attachment_name exceeds the " . ini_get('upload_max_filesize') . 'B limit for the server.';
                break;
              case UPLOAD_ERR_PARTIAL:
                $error = "Only part of the file $attachment_name could be uploaded, please try again.";
                break;
              default:
                $error = "There has been an error attaching the file $attachment_name, please try again.";
            }
            
            return $error;
            
          } // file was uploaded
          
        }
        
        
      }
      
      
      return 'good';
      
    }
    
    // not used yet, but wanna keep code
    public function processSignature()
    {
      
      if(isset($_POST['signature']))
      {
        $sig = $_POST['signature'];
        $sig = str_replace('data:image/png;base64,', '', $sig);
        $sig = str_replace(' ', '+', $sig);
        $data = base64_decode($sig);
        $filename = 'sig-' . date('Y-m-d_H:i:s') . '.png';
        $file = $this->settings['upload_folder'] . $filename;
        
        $this->signature_filename = $filename;
        
        // save sig file?
        if($this->settings['save_signature_files'])
        {
          if( !file_put_contents($file, $data) ){
            return "Sorry, there was an error saving the signature file.";
          }
        }
        
        // add to files array
        //array_push( $this->files, array('temp'=> $file, 'name' => $filename, 'key' => 'signature' ));
        
      }
      
      return 'good';
      
      
    }
    
    public function getMySQLDataHeaders()
    {
      
      $items = $this->settings['mysql_query'];
      
      // ip address?
      if( !empty($this->settings['ip_field']) )
        array_push($items, ['column' => $this->settings['ip_field'] ,'name' => '', 'type' => 'IP']);
      
      // browser?
      if( !empty($this->settings['browser_field']) )
        array_push($items, ['column' => $this->settings['browser_field'],'name' => '', 'type' => 'Browser']);
      
      // date?
      if( !empty($this->settings['date_field']) )
        array_push($items, ['column' => $this->settings['date_field'] ,'name' => '', 'type' => 'Date']);
      
      $headers = '';
      
      foreach($items as $item){
        
        $headers .= '<th>'. $item['column'].'</th>';
        
      }
      
      return $headers;
      
      
    }
    
    
    public function getMySQLDataRows()
    {
      
      $host = $this->settings['mysql_host'];
      $db   = $this->settings['mysql_db'];
      $table = $this->settings['mysql_table'];
      $user = $this->settings['mysql_user'];
      $pass = $this->settings['mysql_password'];
      $charset = 'utf8mb4';
      
      $items = $this->settings['mysql_query'];
      
      // ip address?
      if( !empty($this->settings['ip_field']) )
        array_push($items, ['column' => $this->settings['ip_field'] ,'name' => '', 'type' => 'IP']);
      
      // browser?
      if( !empty($this->settings['browser_field']) )
        array_push($items, ['column' => $this->settings['browser_field'],'name' => '', 'type' => 'Browser']);
      
      // date?
      if( !empty($this->settings['date_field']) )
        array_push($items, ['column' => $this->settings['date_field'] ,'name' => '', 'type' => 'Date']);
      
      
      // Connect via PDO
      $dsn = "mysql:host=$host;dbname=$db;charset=$charset";
      
      $options = [PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION];
      try {
        $pdo = new PDO($dsn, $user, $pass, $options);
        $data = $pdo->query("SELECT * FROM $table")->fetchAll();
        
      } catch (\PDOException $e) {
        return '<b>DB ERROR:</b> <em>'. $sql . "</em><br>" . $e->getMessage();
      }
      
      $pdo = null;
      
      if($data)
      {
        $rows = '';
        foreach ($data as $row){
          $rows .= '<tr>';
          foreach($items as $item){
            $rows .= '<td>';
            
            switch($item['type'])
            {
              case 'File Input':
                if($row[$item['column']])
                  $rows .= '<img src="data:image/jpeg;base64,'.base64_encode( $row[$item['column']] ).'" style="max-height:40px; width:auto"/>';
                break;
                
              case 'Signature':
                $rows .= '<img src="'.$row[$item['column']].'" style="max-height:40px; width:auto"/>';
                break;
                
              case 'Email':
                $rows .= '<a class="text-primary" href="mailto:'.$row[$item['column']].'">'.$row[$item['column']].'</a>';
                break;
                
              default:
                $rows .= $row[$item['column']];
                break;
                
            }
            
            $rows .= '</td>';
            
          }
          
          $rows .= '</tr>';
          
        }
        
      }
      
      return $rows;
      
    }
    
    private function saveToMySQL()
    {
      
      $host = $this->settings['mysql_host'];
      $db   = $this->settings['mysql_db'];
      $user = $this->settings['mysql_user'];
      $pass = $this->settings['mysql_password'];
      $charset = 'utf8mb4';
      
      $data = []; // data for insert
      $values = [];
      $set = '';
      
      // Build query
      //$inserts = '';
      foreach($this->settings['mysql_query'] as $item)
      {
        $name = $item['name'];
        $col = isset($item['column'])?$item['column']:null;
        $type = $item['type'];
        $file_save_as = $item['file_save_as'];
        
        if(isset($_POST['form'][$name]) && !empty($col) ) {
          $data[$col] = $this->value( $_POST['form'][$name], ", ", false );
        }
        
        // save signature dataURL to TEXT column
        if($type == 'Signature')
        {
          $signature_file_path = $this->settings['upload_folder'] . $this->signature_filename;
          
          if( isset($_POST['signature']) && file_exists( $signature_file_path ) ){
            
            switch($file_save_as)
            {
              case 'blob':
                
                $blob = file_get_contents($signature_file_path);
                $data[$col] = addslashes($blob);
                
                break;
                
              case 'url' :
                
                $sig_url = $this->settings['upload_folder_url'].$this->signature_filename;
                $data[$col] = $this->value( $sig_url, ", ", false );
                
                break;
                
              default: // filename
                
                $data[$col] = $this->value( $this->signature_filename , ", ", false );
                
                break;
                
            }
            
            // data string
            //$data[$col] = $this->value( $_POST['form'][$name], ", ", false );
            
          }
        }
        
        // save any files as blob
        if($type == 'File Input')
        {
          
          switch($file_save_as)
          {
            case 'blob':
              
              if( isset($_FILES['fl_'.$name]) && file_exists($_FILES['fl_'.$name]['tmp_name']) ){
                $blob = file_get_contents($_FILES['fl_'.$name]['tmp_name']);
                $data[$col] = addslashes($blob);
              }
              
              break;
              
            case 'url' :
              
              $path = ($this->settings['save_attachment_files']) ? $this->settings['upload_folder_url'].$this->timestamp.'_' :"";
              $data[$col] = $this->value( $path.$_FILES['fl_'.$name]['name'], ", ", false );
              
              break;
              
            default: // filename
              
              $data[$col] = $this->value( $_FILES['fl_'.$name]['name'], ", ", false );
              
              break;
              
          }
          
          
        }
        
        
      }
      
      // ip address?
      if( !empty($this->settings['ip_field']) )
        $data[$this->settings['ip_field']] = $this->getIPAddress();
      
      
      // browser?
      if( !empty($this->settings['browser_field']) )
        $data[$this->settings['browser_field']] = $this->_getBrowser();
      
      
      // date?
      if( !empty($this->settings['date_field']) )
        $data[$this->settings['date_field']] = $this->_getDate();
      
      
      foreach ($data as $key => $value ) {
        $set.="`".str_replace("`", "``", $key)."`". "=:$key, ";
        $values[$key] = $value;
      }
      
      $set = substr($set, 0, -2);
      
      // Connect via PDO
      $dsn = "mysql:host=$host;dbname=$db;charset=$charset";
      
      $options = [PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION];
      try {
        $pdo = new PDO($dsn, $user, $pass, $options);
        $stmt = $pdo->prepare("INSERT INTO `". $this->settings['mysql_db']. "`.`". $this->settings['mysql_table']. "` SET ". $set);
        $stmt->execute($values);
        
      } catch (\PDOException $e) {
        return '<b>DB ERROR:</b> <em>'. $sql . "</em><br>" . $e->getMessage();
      }
      
      $pdo = null;
      
      return 'good';
      
    }
    
    
    
    private function saveToGoogle()
    {
      
      // build add array
      $add = array();
      foreach($this->settings['google_query'] as $item)
      {
        $name = $item['name'];
        $col = isset($item['column'])?$item['column']:null;
        $type = $item['type'];
        $file_save_as = $item['file_save_as'];
        
        if(isset($_POST['form'][$name]) && !empty($col) )
          $add[$col] = $this->value( $_POST['form'][$name], ", ", false );
        
        // save any files
        if($type == 'File Input')
        {
          
          if($file_save_as == 'url')
          {
            
            if( isset($_FILES['fl_'.$name]) && file_exists($_FILES['fl_'.$name]['tmp_name']) ){
              $path = ($this->settings['save_attachment_files']) ? $this->settings['upload_folder_url'].$this->timestamp.'_' :"";
              $add[$col] = $this->value( $path.$_FILES['fl_'.$name]['name'], ", ", false );
            }
            
          }else{
            
            if( isset($_FILES['fl_'.$name]) && file_exists($_FILES['fl_'.$name]['tmp_name']) ){
              $add[$col] = $this->value( $_FILES['fl_'.$name]['name'], ", ", false );
            }
            
          }
          
          
        }
        
        // save signature dataURL to TEXT column
        if($type == 'Signature')
        {
          $signature_file_path = $this->settings['upload_folder'] . $this->signature_filename;
          
          if( isset($_POST['signature']) && file_exists( $signature_file_path ) ){
            
            switch($file_save_as)
            {
                
              case 'url' :
                
                $sig_url = $this->settings['upload_folder_url'].$this->signature_filename;
                $add[$col] = $this->value( $sig_url, ", ", false );
                
                break;
                
              default: // filename
                
                $add[$col] = $this->value( $this->signature_filename , ", ", false );
                
                break;
                
            }
          }
        }
        
      }
      
      // ip address?
      if( !empty($this->settings['ip_field']) )
        $add[$this->settings['ip_field']] = $this->getIPAddress();
      
      
      // browser?
      if( !empty($this->settings['browser_field']) )
        $add[$this->settings['browser_field']] = $this->_getBrowser();
      
      
      // date?
      if( !empty($this->settings['date_field']) )
        $add[$this->settings['date_field']] = $this->_getDate();
      
      
      $rs = $Spreadsheet = new Spreadsheet($this->settings['google_spreadsheet']);
      
      $rs = $Spreadsheet->add( $add );
      
      if($rs != 'good'){
        return $rs;
      }else{
        return 'good';
      }
      
    }
    
    
    
    private function sendMessage($receipt=false)
    {
      
      
      try {
        
        //Create a new PHPMailer instance
        $mail = new PHPMailer();
        $mail->isHTML = true; // make optional ?
        
        
        // Build Mailer
        if($this->settings['use_smtp'])
        {
          
          $mail->isSMTP();
          $mail->SMTPDebug   = $this->settings['smtp_debug'];
          $mail->Debugoutput   = 'html';
          $mail->Host     = $this->settings['smtp_host'];
          $mail->Port     = $this->settings['smtp_port'];
          $mail->SMTPAuth   = $this->settings['smtp_auth'];
          $mail->Username   = $this->settings['smtp_username'];
          $mail->Password   = $this->settings['smtp_password'];
          if($this->settings['smtp_secure'])
            $mail->SMTPSecure   = $this->settings['smtp_secure_prefix'];
          
          
          // Google example:
          //$mail->SMTPAuth = true;                // enable SMTP authentication
          //$mail->SMTPSecure = "tls";              // sets the prefix to the servier
          //$mail->Host = "smtp.gmail.com";        // sets Gmail as the SMTP server
          //$mail->Port = 587;
          // set the SMTP port for the GMAIL
          
        }
        
        
        //  Sign with DKIM
        if($this->settings['dkim'])
        {
          $mail->DKIM_domain = $this->settings['dkim_domain'];
          $mail->DKIM_private = $this->settings['dkim_privkey_path'];
          $mail->DKIM_selector = $this->settings['dkim_selector']; //this effects what you put in your DNS record
          $mail->DKIM_passphrase = $this->settings['dkim_passphrase'];
        }
        
        
        // Reply-To
        // We are sending to this value, not FROM or REPLYTO
        $reply_to_email = $_POST['form'][$this->settings['reply_to_email']];
        $reply_to_name = $_POST['form'][$this->settings['reply_to_name']];
        
        if( !$this->validateEmail($reply_to_email) ) {
          $err = "The value of the Reply-To item ( {$reply_to_email} ) is not a valid email address.";
          @array_push($this->errors, $err);
          unset($err);
        }
        
        // From
        if($this->settings['send_from_reply'] && !$this->settings['reply_using_from'] )
        {
          $from_email = $_POST['form'][$this->settings['reply_to_email']];
          $from_name = $_POST['form'][$this->settings['reply_to_name']];
        }else{
          $from_email = $this->settings['from_email'];
          $from_name = $this->settings['from_name'];
        }
        
        if(!$this->validateEmail($from_email)) {
          $err="The value of the From Email ( ".$from_email." ) is not a valid email address.";
          @array_push($this->errors, $err);
          return $err; // stop here, no need to go further.
        }
        
        
        if($receipt)
        {
          
          // builds receipt subject string
          if( !empty($this->settings['receipt_subject_template']) )
          {
            $subject = $this->parseTemplate('receipt_subject');
            
          }else{
            
            $subject = $this->settings['receipt_prefix'].$_POST['form'][$this->settings['subject']];
            
          }
          
          
          // To senders...
          $to = array($reply_to_email=>$reply_to_name);
          
          // Reply-To
          $mail->AddReplyTo( $from_email, $from_name );
          
          // From
          $mail->SetFrom( $from_email, $from_name );
          // Subject
          $mail->Subject = strip_tags($subject);
          
          $mail->CharSet = $this->settings['encoding_charset'];
          $mail->Encoding = $this->settings['encoding_method'];
          
          $mail->isHTML = true; // make optional ?
          
          
          // Attach Resource File(s) from Owner to Sender ?
          if($this->settings['send_receipt_files']) {
            
            $img_cnt=0;
            
            foreach($this->settings['receipt_files'] as $fileName)
            {
              
              $path = dirname(__DIR__) . DIRECTORY_SEPARATOR .'../../resources/'.$fileName;
              
              if(file_exists($path)){
                
                
                if(@is_array(getimagesize($path))){
                  $mail->AddEmbeddedImage($path, 'image'.$img_cnt++, $fileName);
                }else{
                  $mail->addAttachment( $path );
                }
                
                
              }
            }
          }
          
          // Body
          $mail->msgHTML($this->receipt_content);
          // Add Plain Text Part
          $mail->AltBody = $this->_html2Text($this->receipt_content);
          
          
        }else{
          
          // builds email subject string
          
          if( !empty($this->settings['email_subject_template']) )
          {
            $subject = $this->parseTemplate('email_subject');
            
          }else{
            
            $subject = $_POST['form'][$this->settings['subject']];
            
          }
          
          $to = $this->settings['to'];
          
          // Cc recipients...
          if($this->settings['cc']) {
            $cc = $this->settings['cc'];
            foreach ($cc as $address => $name)
            {
              if (is_int($address)) {
                $mail->AddCC($name);
              } else {
                $mail->AddCC($address , $name);
              }
            }
          }
          
          // Bcc recipients...
          if($this->settings['bcc']) {
            $bcc = $this->settings['bcc'];
            foreach ($bcc as $address => $name)
            {
              if (is_int($address)) {
                $mail->AddBCC($name);
              } else {
                $mail->AddBCC($address , $name);
              }
            }
          }
          
          // ReplyTo recipients...
          // these are in addition to the dynamic value sent from form.
          
          // primary reply-to (dynamic)
          $mail->AddReplyTo( $reply_to_email, $reply_to_name );
          
          if($this->settings['reply']) {
            $reply = $this->settings['reply'];
            foreach ($reply as $address => $name)
            {
              if (is_int($address)) {
                $mail->AddReplyTo($name);
              } else {
                $mail->AddReplyTo($address , $name);
              }
            }
          }
          
          
          
          // From (static value, not dynamic )
          $mail->SetFrom( $from_email, $from_name );
          
          // Subject
          $mail->Subject = strip_tags($subject);
          
          $mail->CharSet = $this->settings['encoding_charset'];
          $mail->Encoding = $this->settings['encoding_method'];
          
          
          // Attachments?
          if($this->files){
            for ( $i = 0; $i < sizeof ( $this->files ); $i++ )
            {
              
              if(@is_array(getimagesize($this->files[$i]['temp']))){
                
                $mail->AddEmbeddedImage($this->files[$i]['temp'], 'image_attach'.$i, $this->files[$i]['name'] );
                
              }else{
                
                $mail->addAttachment($this->files[$i]['temp'], $this->files[$i]['name']);
                
              }
              
              
            }
          }
          
          // Body
          $mail->msgHTML($this->email_content);
          // Add Plain Text Part
          $mail->AltBody = $this->_html2Text($this->email_content);
          
          
        }
        
        
        
        // Send the message
        $failed = array();
        $numSent = 0;
        
        //send the message, check for errors
        
        if(is_array($to)) {
          foreach ($to as $address=>$name)
          {
            if (is_int($address)) {
              $mail->AddAddress($name);
            } else {
              $mail->AddAddress($address, $name);
            }
            //send the message, check for errors
            if (!$mail->send()) {
              $failed[] .= $mail->ErrorInfo;
            } else {
              $numSent++;
            }
            
            $mail->ClearAddresses();
          }
        }
        
        
        if(!$numSent || count($failed) )
        {
          $error = "$numSent message were sent. The following addresses had problems, ".implode(", ", $failed);
          
          @array_push($this->errors, $error);
          
          return;
          
        }
        
        
        
      } catch (phpmailerException $e) {
        echo $e->errorMessage(); //Pretty error messages from PHPMailer
      } catch (Exception $e) {
        echo $e->getMessage(); //Boring error messages from anything else!
      }
      
      $mail->ClearAllRecipients( ); // clear all
      
      unset($mail);
      unset($message);
      
      return 'good';
      
    }
    
    private function buildEmailContent()
    {
      
      $tpl = trim($this->settings['email_template']);
      
      if( empty($tpl) || !$this->settings['email_custom_content'] )
      {
        
        $this->email_content = $this->buildDefaultHtmlContent();
        
      }else{
        
        $this->email_content = $this->parseTemplate('email');
        
      }
      
    }
    
    private function buildReceiptContent()
    {
      
      $tpl = trim($this->settings['receipt_template']);
      
      if( empty($tpl) || !$this->settings['receipt_custom_content'] )
      {
        $this->receipt_content = $this->buildDefaultHtmlContent('receipt');
      }else{
        $this->receipt_content = $this->parseTemplate('receipt');
      }
      
    }
    
    //=====================================================//
    // buildDefaultHtmlContent
    //=====================================================//
    
    private function buildDefaultHtmlContent($type='email')
    {
      
      
      
      $content= '<table style="width:100%; margin:20px 0;"><tbody>';
      $images='<table style="width:100%;"><tbody>';
      $signature='';
      
      
      if(isset($_POST['signature']) && !empty($_POST['signature']) )
      {
        $signature='<table style="width:100%; margin:8px 4px;"><tbody>';
        $signature .= '<tr><td class="sig-img" style="color:#cccccc; text-align:center; padding:4px; font-size:11px;">SIGNATURE:<br><img src="'.$_POST['signature'].'"></td></tr>';
        $signature .= '</tbody></table>';
      }
      
      
      
      if($type == 'receipt')
      {
        
        if(isset($this->settings['tpl_receipt_txt']))
          $content .= '<h3 class="headline" style="text-align:center;">'.$this->settings['tpl_receipt_txt'].'</h3><hr>';
        
        $content .= '<table style="width:100%;margin-top: 20px\">'.PHP_EOL;
        $title = ( !empty($this->settings['receipt_subject_template']) )?$this->parseTemplate('receipt_subject'):$this->settings['receipt_prefix'].$_POST['form'][$this->settings['subject']];
        
        // add embedded images
        if($this->settings['send_receipt_files']) {
          
          $img_cnt=0;
          
          foreach($this->settings['receipt_files'] as $fileName)
          {
            
            $path = dirname(__DIR__) . DIRECTORY_SEPARATOR .'../../resources/'.$fileName;
            
            if( file_exists($path) && @is_array(getimagesize($path)) ){
              
              $images .= '<tr><td><hr></td></tr>';
              
              $images .= '<tr><td class="attachment"><img src="cid:image'.$img_cnt++.'"></td></tr>';
              
            }
          }
        }
        
      }else{
        
        if($this->files){
          for ( $i = 0; $i < sizeof ( $this->files ); $i++ )
          {
            
            if(@is_array(getimagesize($this->files[$i]['temp']))){
              
              $images .= '<tr><td><hr></td></tr>';
              
              $images .= '<tr><td class="attachment"><img src="cid:image_attach'.$i.'"><td></tr>';
              
            }
            
          }
        }
        
      }
      
      $images .= '</tbody></table>';
      
      
      $form = $_REQUEST['form']; // need to use REQUEST because POST has issues with line-breaks
      
      $subject = $form[$this->settings['subject']];
      
      if($type == 'receipt')
        $subject =  $this->settings['receipt_prefix'].$subject;
      
      //build a message from the reply for both HTML and text in one loop.
      foreach ($form as $key => $value) {
        
        $label = $this->settings['form_fields'][$key];
        $content .= '<tr><td class="label">' . strip_tags($label) . '</td><td class="value"> ';
        $content .= nl2br($this->value($value)) . "</td></tr>".PHP_EOL;
        
      }
      //close the HTML content.
      $content .= '</tbody></table>'.PHP_EOL;
      $content .= '</body></html>';
      
      $subject = ( !empty($this->settings['email_subject_template']) )?$this->parseTemplate('email_subject'):$_POST['form'][$this->settings['subject']];
      
      
      $logo = ($this->settings['logo_img'] && $this->settings['tpl_show_logo'] ) ? '<img style="width:160px; max-width: 100%; height:auto;" src="'.$this->settings['logo_img'].'">' : '';
      
      $footer = (isset($this->settings['tpl_footer_txt'])) ? $this->settings['tpl_footer_txt'] : '';
      
      if( ( $type == 'email' && !$this->settings['email_use_wrapper'] ) || ( $type == 'receipt' && !$this->settings['receipt_use_wrapper'] ) )
      {
        return $content;
        
      }else{
        
        $str = file_get_contents(dirname(__FILE__) . DIRECTORY_SEPARATOR .'mailTemplate.php');
      }
      
      
      $browser = ($this->settings['tpl_show_browser'])?$this->_getBrowser():"";
      $ip = ($this->settings['tpl_show_ip'])?$this->getIPAddress():"";
      $date = ($this->settings['tpl_show_date'])?$this->_getDate():"";
      $parent = (isset($_REQUEST['yd_url']) && $this->settings['tpl_show_parent_url'])?$_REQUEST['yd_url']:"";
      $title = ($this->settings['tpl_show_subject'])?$this->_safeEscapeString($subject):"";
      
      $str = str_replace('%%title%%', $title, $str);
      $str = str_replace('%%hexcode%%', '#226AB6', $str);
      $str = str_replace('%%logo%%', $logo, $str);
      $str = str_replace('%%content%%', $content , $str);
      $str = str_replace('%%footer%%', $footer , $str);
      $str = str_replace('%%images%%', $images , $str);
      $str = str_replace('%%date%%', $date , $str);
      $str = str_replace('%%browser%%', $browser , $str);
      $str = str_replace('%%ip%%', $ip , $str);
      $str = str_replace('%%signature%%', $signature , $str);
      $str = str_replace('%%css%%', '', $str);
      $str = str_replace('%%demo%%', $this->_demoString() , $str);
      $str = str_replace('%%parent%%', $parent, $str);
      
      return $str;
      
    }
    
    //=====================================================//
    // parseTemplate
    //=====================================================//
    
    
    private function parseTemplate($type = 'email')
    {
      
      switch($type)
      {
        case 'email':
          $content =  $this->settings['email_template'];
          $subject = ( !empty($this->settings['email_subject_template']) )?$this->parseTemplate('email_subject'):$_POST['form'][$this->settings['subject']];
          $encode = true;
          break;
          
        case 'email_subject':
          $content =  $this->settings['email_subject_template'];
          $encode = false;
          break;
          
        case 'receipt':
          $content =  $this->settings['receipt_template'];
          $subject = ( !empty($this->settings['receipt_subject_template']) )?$this->parseTemplate('receipt_subject'):$this->settings['receipt_prefix'].$_POST['form'][$this->settings['subject']];
          $encode = true;
          break;
          
        case 'receipt_subject':
          $content =  $this->settings['receipt_subject_template'];
          $encode = false;
          break;
          
        default:
          $content =  $this->settings['email_template'];
          break;
      }
      
      $form = $_REQUEST['form']; // need to use REQUEST because POST has issues with line-breaks
      
      foreach ($form as $key => $value)
      {
        $content = str_replace("#{$key}#", nl2br($this->value($value, ", ", $encode)), $content);
      }
      
      // browser
      $agent = $this->browser->getBrowser().' v.'.$this->browser->getVersion(). ' ('.$this->browser->getPlatform().')';
      $content = str_replace("#fl-browser#", $this->_getBrowser() , $content);
      // ip address
      $content = str_replace("#fl-ip_address#", $this->getIPAddress() , $content);
      $content = str_replace("#fl-ip#", $this->getIPAddress() , $content);
      // date
      $content = str_replace("#fl-date#", $this->_getDate() , $content);
      // parent page
      $content = (isset($_REQUEST['url'])) ? str_replace("#fl-page#", $_REQUEST['url'], $content) : str_replace("#fl-page#", "", $content);
      
      // Get rid of any unparsed place-holders
      $content = preg_replace('/#([A-Za-z0-9-._]*)#/', '', $content);
      
      //
      if($type == 'receipt_subject' || $type == 'email_subject')
        return $content;
      
      $logo = ( $this->settings['logo_img'] && $this->settings['tpl_show_logo'] ) ? '<img style="width:160px; max-width: 100%; height:auto;" src="'.$this->settings['logo_img'].'">' : '';
      $footer = (isset($this->settings['tpl_footer_txt'])) ? $this->settings['tpl_footer_txt'] : '';
      
      if(  $type == 'receipt_subject' || $type == 'email_subject' || ( $type == 'email' && !$this->settings['email_use_wrapper'] ) || ( $type == 'receipt' && !$this->settings['receipt_use_wrapper'] ) )
      {
        
        return $content;
        
      }else{
        
        $str = file_get_contents(dirname(__FILE__) . DIRECTORY_SEPARATOR .'mailTemplate.php');
      }
      
      $signature='';
      if(isset($_POST['signature']) && !empty($_POST['signature']) )
      {
        $signature='<table style="width:100%; margin:8px 4px;"><tbody>';
        $signature .= '<tr><td class="sig-img" style="color:#cccccc; text-align:center; padding:4px; font-size:11px;">SIGNATURE:<br><img src="'.$_POST['signature'].'"></td></tr>';
        $signature .= '</tbody></table>';
      }
      
      $browser = ($this->settings['tpl_show_browser'])?$this->_getBrowser():"";
      $ip = ($this->settings['tpl_show_ip'])?$this->getIPAddress():"";
      $date = ($this->settings['tpl_show_date'])?$this->_getDate():"";
      $parent = (isset($_REQUEST['yd_url']) && $this->settings['tpl_show_parent_url'])?$_REQUEST['yd_url']:"";
      $title = ($this->settings['tpl_show_subject'])?$this->_safeEscapeString($subject):"";
      
      $str = str_replace('%%title%%', $title, $str);
      $str = str_replace('%%hexcode%%', '#226AB6', $str);
      $str = str_replace('%%logo%%', $logo, $str);
      $str = str_replace('%%content%%', $content , $str);
      $str = str_replace('%%footer%%', $footer , $str);
      $str = str_replace('%%images%%', $images , $str);
      $str = str_replace('%%date%%', $date , $str);
      $str = str_replace('%%browser%%', $browser , $str);
      $str = str_replace('%%ip%%', $ip , $str);
      $str = str_replace('%%signature%%', $signature , $str);
      $str = str_replace('%%css%%', '', $str);
      $str = str_replace('%%demo%%', $this->_demoString() , $str);
      $str = str_replace('%%parent%%', $parent, $str);
      
      
      return $str;
      
    }
    
    
    //=====================================================//
    // getOutput
    //=====================================================//
    
    public function getOutput($errors, $required) {
      
      $errors = array_merge($errors, $this->errors);
      $errors = array_unique($errors);
      $msg = ''; // init
      
      if(count($errors)){
        $missing = array();
        foreach($errors as $error)
        $msg .= '<br>'.str_replace('"', '\'', $error);
        
        if($this->settings['error_redirect'])
        {
          return json_encode( array('result'=>'fail', 'msg'=>$this->settings['redirect_message'],'redirect'=>1, 'redirect_url'=>$this->settings['error_url'] ) );
        }
        
        if(isset($required) && is_array($required))
        {
          $missing = $required;
        }
        return json_encode( array('result'=>'fail', 'msg'=>$msg, 'required'=>$missing, 'redirect'=>0, 'redirect_url'=>'' ) );
        unset($msg);
        unset($missing);
      }else{
        
        $msg = $this->settings['success_message'];
        
        if($this->settings['success_redirect'])
        {
          return json_encode( array('result'=>'good', 'msg'=>$msg, 'redirect'=>1, 'redirect_url'=>$this->settings['success_url']) );
        }else{
          return json_encode( array('result'=>'good', 'msg'=>$msg, 'redirect'=>0, 'redirect_url'=>'' ) );
        }
      }
      
      return;
      
      
    }
    
    
    //=====================================================//
    // validateEmail
    //=====================================================//
    
    public function validateEmail($email) {
      //check for all the non-printable codes in the standard ASCII set,
      //including null bytes and newlines, and exit immediately if any are found.
      if (preg_match("/[\\000-\\037]/",$email)) {
        return false;
      }
      $pattern = "/^[-_a-z0-9\'+*$^&%=~!?{}]++(?:\.[-_a-z0-9\'+*$^&%=~!?{}]+)*+@(?:(?![-.])[-a-z0-9.]+(?<![-.])\.[a-z]{2,16}|\d{1,3}(?:\.\d{1,3}){3})(?::\d++)?$/iD";
      if(!preg_match($pattern, $email)){
        return false;
      }
      return true;
    }
    
    //=====================================================//
    // checkRequiredFields
    //=====================================================//
    
    // Function to check the required fields are filled in.
    public function checkRequiredFields() {
      
      $err = 0;
      $fields = array();
      $rf = array(); // required fields array
      
      // loops through all post args and look for required args...
      foreach($_POST as $key => $value) {
        if(preg_match('/_required/', $key, $matches ))
        {
          array_push($rf, str_replace('_required', '', $key));
        }
      }
      
      foreach($rf as $key) {
        
        if (isset($_FILES[$key])) {
          $string = (isset($_FILES[$key]['name'])) ? $_FILES[$key]['name'] : '';
        } else {
          $string = $this->value($_POST['form'][$key]);
        }
        
        if( empty($string) )
        {
          array_push($fields, $key);
          $err++;
        }
        
        unset($v);
        
      }
      
      if($err){
        return $fields;
      }else{
        return 'good';
      }
    }
    
    //=====================================================//
    // validateUserLogin
    //=====================================================//
    
    function validateUserLogin($name, $password)
    {
      
      $users = $this->settings['user_logins'];
      
      foreach ( $users as $user => $pw ) {
        
        if( $name === $user && $password === $pw ){
          return true;
        }else{
          $ip = $this->getIPAddress();
          $msg = "Failed Login Attempt!\nUser:{$name}\nPassword:{$password}\nIP:{$ip}\n";
          $this->notifyOwner($msg);
        }
        
      }
      
      return false;
      
    }
    
    //=====================================================//
    // notifyOwner
    //=====================================================//
    
    function notifyOwner($message)
    {
      // notify owner?
      if($this->settings['notify_enabled']){
        $headers = 'Content-type: text/plain; charset=utf-8' . PHP_EOL;
        $headers .= 'From: "'.$this->settings['from_name'].'" <'.$this->settings['from_email'].'>' . PHP_EOL;
        $notification = "FormLoom Notification:".PHP_EOL.PHP_EOL;
        $notification .= "Page: ".$_SERVER['HTTP_REFERER']. PHP_EOL.PHP_EOL;
        $notification .= "Date: ".date('r'). PHP_EOL.PHP_EOL;
        $notification .= "Error: ".PHP_EOL.PHP_EOL.$message;
        @mail($this->settings['notify_email'], 'FormLoom Notification', $notification, $headers);
      }
      
    }
    
    
    //=====================================================//
    // value
    //=====================================================//
    
    public function value($var, $sep=", ", $encode=true){
      
      if($encode)
        $var = $this->_htmlEntities($var); // encode all html tags
      
      if(is_array($var)){
        $str = (isset($var)) ? implode($sep, $var) : '';
      }else{
        $str = (isset($var)) ? $var : '';
      }
      
      // added stringCleaner in 3.0.4 - 01-02-2015 09:27:18 am -MY
      return $this->_stringCleaner($str);
    }
    
    //=====================================================//
    // stringCleaner
    //=====================================================//
    
    private function _stringCleaner($str) {
      
      // removes smart quotes ( and such ) from strings and converts them to ascii
      
      $quotes = array(
                      "\xC2\xAB"     => '"', // « (U+00AB) in UTF-8
                      "\xC2\xBB"     => '"', // » (U+00BB) in UTF-8
                      "\xE2\x80\x98" => "'", // ‘ (U+2018) in UTF-8
                      "\xE2\x80\x99" => "'", // ’ (U+2019) in UTF-8
                      "\xE2\x80\x9A" => "'", // ‚ (U+201A) in UTF-8
                      "\xE2\x80\x9B" => "'", // ‛ (U+201B) in UTF-8
                      "\xE2\x80\x9C" => '"', // “ (U+201C) in UTF-8
                      "\xE2\x80\x9D" => '"', // ” (U+201D) in UTF-8
                      "\xE2\x80\x9E" => '"', // „ (U+201E) in UTF-8
                      "\xE2\x80\x9F" => '"', // ‟ (U+201F) in UTF-8
                      "\xE2\x80\xB9" => "'", // ‹ (U+2039) in UTF-8
                      "\xE2\x80\xBA" => "'", // › (U+203A) in UTF-8
                      );
      
      $str = strtr($str, $quotes);
      
      return $str;
      
    }
    
    private function _demoString()
    {
      if($this->settings['registered'])
        return '';
      
      return 'SENT WITH UNREGISTERED VERSION OF <strong><a href="https://yabdab.com/plugins/formloom" style="color:#316DD8;">FORMLOOM 4</a></strong>';
    }
    
    //=====================================================//
    // safeEscapeString
    //=====================================================//
    
    // Function to escape data inputted from users. This is to protect against embedding
    // of malicious code being inserted into the HTML email.
    
    private function _safeEscapeString($string) {
      $str = strip_tags($string);
      $str = $this->_htmlEntities($str);
      return $str;
    }
    
    //=====================================================//
    // htmlEntities
    //=====================================================//
    
    private function _htmlEntities($mixed, $quote_style = ENT_QUOTES, $charset = 'UTF-8')
    {
      if (is_array($mixed)) {
        foreach($mixed as $key => $value) {
          $mixed[$key] = $this->_htmlEntities($value, $quote_style, $charset);
        }
      } elseif (is_string($mixed)) {
        $mixed = htmlentities(html_entity_decode($mixed, $quote_style), $quote_style, $charset);
      }
      return $mixed;
    }
    
    //=====================================================//
    // html2Text
    //=====================================================//
    
    private function _html2Text($str)
    {
      $text_str = str_replace("<br />", PHP_EOL, $str); // convert breaks into line-breaks.
      $text_str = strip_tags($text_str); // remove all markup.
      $text_str = htmlspecialchars_decode($text_str); // convert values with special chars back to unencoded.
      return $text_str;
    }
    
    //=====================================================//
    // http_post
    //=====================================================//
    
    private function _http_post($params)
    {
      ;
      $postData = ''; // init
      
      // add files into the params first...
      
      // Attachments?
      foreach ($_FILES as $param => $file) {
        $params[$param] = '@' . $file['tmp_name'] . ';filename=' . $file['name'] . ';type=' . $file['type'];
      }
      
      // Attach Resource File ?
      if($this->settings['receipt_attach_file'] && file_exists($this->settings['receipt_file']))
      {
        
        //$file = array('receipt_file' => new CurlFile($this->settings['receipt_file'], '' /* MIME-Type */, ''));
        //array_push($params, $file);
      }
      
      
      $ch = curl_init();
      $header = array("Content-type: multipart/form-data");
      curl_setopt($ch,CURLOPT_URL,$url);
      curl_setopt($ch,CURLOPT_USERAGENT, "Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)");
      curl_setopt($ch,CURLOPT_RETURNTRANSFER,TRUE); // false seems to return errors from woosh script
      curl_setopt($ch,CURLOPT_HTTPHEADER,$header);
      curl_setopt($ch,CURLOPT_ENCODING,"");
      curl_setopt($ch,CURLOPT_FOLLOWLOCATION, TRUE);
      curl_setopt($ch, CURLOPT_POST, TRUE);
      curl_setopt($ch, CURLOPT_POSTFIELDS, $params);
      
      $output=curl_exec($ch);
      
      if($output === false)
      {
        echo "Error Number:".curl_errno($ch)."<br>";
        echo "Error String:".curl_error($ch);
      }
      
      curl_close($ch);
      return $output;
      
    }
    
    
    private function _getBrowser()
    {
      return $this->browser->getBrowser().' v.'.$this->browser->getVersion(). ' ('.$this->browser->getPlatform().')';
    }
    
    public function getIPAddress()
    {
      $ip = (empty ( $_SERVER ['HTTP_CLIENT_IP'] ) ? (empty ( $_SERVER ['HTTP_X_FORWARDED_FOR'] ) ? $_SERVER ['REMOTE_ADDR'] : $_SERVER ['HTTP_X_FORWARDED_FOR']) : $_SERVER ['HTTP_CLIENT_IP']);
      return $this->_anonymizeIP($ip);
    }
    
    private function _getDate()
    {
      $date = ($this->settings['date_format']) ? date($this->settings['date_format']) : date("Y-m-d H:i:s");
      return $date;
    }
    
    private function _anonymizeIP($ip) {
      
      // string IPv4 netmask used to anonymize IPv4 address.
      $ipv4NetMask = "255.255.255.0";
      
      // string IPv6 netmask used to anonymize IPv6 address.
      $ipv6NetMask = "ffff:ffff:ffff:ffff:0000:0000:0000:0000";
      
      $packedAddress = @inet_pton($ip);
      
      if (strlen($packedAddress) == 4) {
        return inet_ntop(inet_pton($ip) & inet_pton($ipv4NetMask));
      } elseif (strlen($packedAddress) == 16) {
        return inet_ntop(inet_pton($ip) & inet_pton($ipv6NetMask));
      }
      
      return '0.0.0.0';
      
    }
    
    
    
    
  }
  
  //=====================================================//
  // !  THE END
  //=====================================================//
